#ifdef FCACHE_H_INCLUDED
#else
#define FCACHE_H_INCLUDED

#define NTCP 10

#define CACHEFORWARDADDRESSLENGTH 4096
#include "glite/data/io/daemon/protocol.hxx"
#include "glite/data/io/daemon/aioerror.hxx"
#include <sys/shm.h>
#include "glite/data/io/quanta/QUANTA.hxx"
#include <string.h>
#include <signal.h>

#define MAXCHUNKS 20
#define DEFAULTAIODPORT 9999
#define _PERFORMANCEDEBUG

#define DEFAULTMONITORPORT 10000
// information lookup service
#define DO_MONITORINFORMATION 0xaaaa
#define DO_CACHELOOKUP        0xaabb

////////////////////////////////////////////////////////////////////////////
// this structure has to be 32bit aligned ... will be changed ..

struct connectionblock {
  int cryptauthentication;
  int cryptdata;
  int cachelevel;
  long long requestsize; // used for write's ....
  int permissions;
  char serverid;
  char lcachehopcnt;
  char rcachehopcnt;
  char mcachehopcnt;
  char cachehopcnt;
  char cacheforwardaddress[CACHEFORWARDADDRESSLENGTH];
  char login[256];
  char protocol[256];
  char lfn[FILENAMELENGTH];
  char guid[40];
  char Protocol_Number[PROTOCOL_NUMBERLENGTH];         //Protocol Number
} ;

struct connectionreturnblock {
  char remotefilename[FILENAMELENGTH];
  char guid[40];
  char lfn[FILENAMELENGTH];
  //char reason[ERROR_REASONLENGTH];                  //error reason
} ;

struct filecachechunk {
    int blocknr;
    char* ptr;
    int   size;
    int   locker;
} ;

struct filecacheinfo {
    int shm_id;
    char filename[FILENAMELENGTH];
    char diskcachedir[1024];
    long diskcachesize;
    char hostname[255];
    int chunksize;
    int ntcp;
    int nchunks;
    int   remoteport;
    int remoteaccess;
    char iopluginname[1024];
    char iopluginhelpname[1024];
    struct connectionblock sslcrypt;
    sig_atomic_t * run;
} ;

struct filecachestat {
	
	int32   dev;         /* device */
	int64   ino;         /* inode */
	int32 	mode;        /* protection */
	int32   nlink;       /* number of hard links */
	int32 	uid;         /* user ID of owner */
	int32 	gid;         /* group ID of owner */
	int32   rdev;        /* device type (if inode device) */
	int64   size;        /* total size, in bytes */
	int32   blksize;     /* blocksize for filesystem I/O */
	int64   blocks;      /* number of blocks allocated */
	int32	atime;       /* time of last access */
	int32   mtime;       /* time of last modification */
	int32   ctime;       /* time of last change */
    char 	owner[1024]; /* name of owner */
	char 	group[1024]; /* name of group */
};

// if we include fcache.hxx from libalienio.hxx, we want only the structures !
#ifndef FCACHE_TYPES_ONLY 


////////////////////////////////////////////////////////////////////////////
int shm_exists(int key)
{
  int result;
  result=shmget(key,0,IPC_EXCL|SHM_W|SHM_R);
  if(result==-1) return 0; else return 1;
}

/* create shared memory block and attach it */

char *shm_new(int key,int size)
{
  int result;
  void *p;
  result=shmget(key,size,SHM_W|SHM_R|IPC_CREAT|IPC_EXCL);
  if(result==-1) return (char *) -1;
  p=shmat(result,0,0);
  return (char*)p;
}

int shm_destroy(int key)
{
  int result;
  result=shmget(key,0,SHM_W|SHM_R|IPC_EXCL);
  shmctl(result,IPC_RMID,0);
  return 0;
}

char *shm_attach(int key)
{
  int result;
  char *p;
  result=shmget(key,0,SHM_W|SHM_R|IPC_EXCL);
  p=(char*)shmat(result,0,0);
  return (char*)p;
}

int shm_detach(char *p)
{
  return (int) shmdt(p);
}

void extractfromcacheforward(const char* cacheforward, char* newhostname, int* newport) {
    /////////////////////////////////////////////////////////
    // addresses are like:
    // <host1>:<port1>@<host2>:<port2>@ ....<hostn>:<portn>
    // ( on host n is our destination file ...
    /////////////////////////////////////////////////////////
    char hostportname[1024];
    char hostname[1024];
    char portname[1024];
    char *eofhostname;
    char *porthostname;

    // take the 1st hostname out of the cacheforwardaddresses
    eofhostname = strchr(cacheforward,'@');
    
    if (!eofhostname)
	// or take the only hostname ....
	eofhostname = (char*)cacheforward + strlen(cacheforward);
    
    /////////////////////////////////////////////////////////
    // take out the relevant part ....
    // .. should be like <host>:<port>
    memcpy(hostportname,(cacheforward), eofhostname - (cacheforward));
    hostportname[eofhostname - (cacheforward)] = 0;
    porthostname = strchr(hostportname,':');
    if (porthostname) {
	memcpy(hostname,hostportname, porthostname - hostportname);
	hostname[porthostname-hostportname] = 0;
	memcpy(portname,porthostname+1, strlen(hostportname) - (porthostname + 1 - hostportname));
	portname[strlen(hostportname) - (porthostname + 1 - hostportname)] = 0;
	*newport = atoi(portname);
    } else {
	memcpy(hostname,hostportname,strlen(hostportname));
	hostname[strlen(hostportname)] = 0;
    }
    sprintf(newhostname,"%s",hostname);
}


///////////////////////////////////////////////////////////////////////
// connectionblock struct streamer
///////////////////////////////////////////////////////////////////////
int writeconnectionblock(QUANTAnet_extendedParallelTcpClient_c *client,struct connectionblock* cblock) {
    if (!cblock)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    int32 s_cryptauthentication = cblock->cryptauthentication;
    int32 s_cryptdata           = cblock->cryptdata;
    int32 s_cachelevel          = cblock->cachelevel;
    int64 s_requestsize         = cblock->requestsize;
    int32 s_permissions         = cblock->permissions;

    if ( (client->writeInt32(s_cryptauthentication)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeInt32(s_cryptdata)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeInt32(s_cachelevel)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeInt64(s_requestsize)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeInt32(s_permissions)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeChar(cblock->serverid)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;
    
    if ( (client->writeChar(cblock->lcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeChar(cblock->rcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeChar(cblock->mcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->writeChar(cblock->cachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    int nwrite;

    nwrite = CACHEFORWARDADDRESSLENGTH;
    if ( (client->write(cblock->cacheforwardaddress,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    nwrite = 256;
    if ( (client->write(cblock->login,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    nwrite = 256;
    if ( (client->write(cblock->protocol,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;


    //protocol_number added to connectionnblock
    nwrite = 256;
    if ( (client->write(cblock->Protocol_Number,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;



  
    if(0 != cblock->cryptauthentication){
        char * filenamebuffer;
        int filenamebuffer_len; 
        // crypt the filename
        int r = client->getContext()->encrypt(cblock->lfn,strlen(cblock->lfn)+1,filenamebuffer,filenamebuffer_len,QUANTAnet_auth_context_c::PM_PRIVACY);
        if (0 != r){
            return AIO_CRYPTENVERROR;
        }
        nwrite = filenamebuffer_len;
        if ( (client->writeInt32(nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED){
            client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
            return QUANTAnet_parallelTcpClient_c::FAILED;
        }
        if ( (client->write(filenamebuffer,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED){
            client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
            return QUANTAnet_parallelTcpClient_c::FAILED;
        }
        client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
        
        // crypt the guid
        r = client->getContext()->encrypt(cblock->guid,strlen(cblock->guid)+1,filenamebuffer,filenamebuffer_len,QUANTAnet_auth_context_c::PM_PRIVACY);
        if (0 != r){
            return AIO_CRYPTENVERROR;
        }
        nwrite = filenamebuffer_len;
        if ( (client->writeInt32(nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED){
            client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
            return QUANTAnet_parallelTcpClient_c::FAILED;
        }
        if ( (client->write(filenamebuffer,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED){
            client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
            return QUANTAnet_parallelTcpClient_c::FAILED;
        }
        client->getContext()->releaseBuffer(filenamebuffer,filenamebuffer_len);
    } else{
        // Lfn
        nwrite = strlen(cblock->lfn) + 1;
        if ( (client->writeInt32(nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        if ( (client->write(cblock->lfn,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        nwrite = strlen(cblock->guid) + 1;
        // Guid
        if ( (client->writeInt32(nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        if ( (client->write(cblock->guid,&nwrite)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    return QUANTAnet_parallelTcpClient_c::OK;
}

int readconnectionblock (QUANTAnet_extendedParallelTcpClient_c *client,struct connectionblock* cblock) {

    if ( (client->readInt32(cblock->cryptauthentication)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readInt32(cblock->cryptdata)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readInt32(cblock->cachelevel)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;
  
    if ( (client->readInt64(cblock->requestsize)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readInt32(cblock->permissions)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readChar(cblock->serverid)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readChar(cblock->lcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readChar(cblock->rcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readChar(cblock->mcachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    if ( (client->readChar(cblock->cachehopcnt)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    int nread;
    nread = CACHEFORWARDADDRESSLENGTH;
    if ( (client->read(cblock->cacheforwardaddress,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    nread = 256;
    if ( (client->read(cblock->login,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;

    nread = 256;
    if ( (client->read(cblock->protocol,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;
  
    //Protocol number added to connectionblock
    nread = 256;
    if ( (client->read(cblock->Protocol_Number,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
        return QUANTAnet_parallelTcpClient_c::FAILED;


    if(0 != cblock->cryptauthentication){
        nread = 0;
        char * filenamebuffer;
        int filenamebuffer_len; 
        char * dstbuffer     = 0;
        int    dstbuffer_len = 0;
    
        // Read LFN
        if ( (client->readInt32(filenamebuffer_len)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        filenamebuffer = new char[filenamebuffer_len];    
        if ( (client->read(filenamebuffer,&filenamebuffer_len)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        int r = client->getContext()->decrypt(filenamebuffer,filenamebuffer_len,dstbuffer,dstbuffer_len);
        if (0 != r){
           delete [] filenamebuffer;
           return AIO_DECRYPTENVERROR;
        }
        strncpy(cblock->lfn,dstbuffer,(dstbuffer_len < FILENAMELENGTH) ? dstbuffer_len : FILENAMELENGTH);
        client->getContext()->releaseBuffer(dstbuffer,dstbuffer_len);
        delete [] filenamebuffer;
        
        // Read GUID
        if ( (client->readInt32(filenamebuffer_len)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        filenamebuffer = new char[filenamebuffer_len];    
        if ( (client->read(filenamebuffer,&filenamebuffer_len)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        r = client->getContext()->decrypt(filenamebuffer,filenamebuffer_len,dstbuffer,dstbuffer_len);
        if (0 != r){
           delete [] filenamebuffer;
           return AIO_DECRYPTENVERROR;
        }
        strncpy(cblock->guid,dstbuffer,(dstbuffer_len < 40)? dstbuffer_len : 40);
        client->getContext()->releaseBuffer(dstbuffer,dstbuffer_len);
        delete [] filenamebuffer;
    } else {
        // Read LFN
        nread = 0;
        if ( (client->readInt32(nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        if ( (client->read(cblock->lfn,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        // Read GUID
        nread = 0;
        if ( (client->readInt32(nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
        if ( (client->read(cblock->guid,&nread)) == QUANTAnet_parallelTcpClient_c::FAILED)
            return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    return QUANTAnet_parallelTcpClient_c::OK;
}

int writefilecachestat(QUANTAnet_extendedParallelTcpClient_c *client,struct filecachestat * file_stats) {
    if (!file_stats){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write Device
    if ((client->writeInt32(file_stats->dev))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write iNode
    if ((client->writeInt64(file_stats->ino))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write Mode
    if ((client->writeInt32(file_stats->mode))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write Number of Hard Link
    if ((client->writeInt32(file_stats->nlink))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write OwnerId
    if ((client->writeInt32(file_stats->uid))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write GroupId
    if ((client->writeInt32(file_stats->gid))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write Device Type
    if ((client->writeInt32(file_stats->rdev))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write Size
    if ((client->writeInt64(file_stats->size))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write Block Size
    if ((client->writeInt32(file_stats->blksize))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write Block Count
    if ((client->writeInt64(file_stats->blocks))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write Last Access Time
    if ((client->writeInt32(file_stats->atime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write Last Modification Time
    if ((client->writeInt32(file_stats->mtime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Write Last Change Time
    if ((client->writeInt32(file_stats->ctime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Write File Owner
    int owner_length = strlen(file_stats->owner) + 1;
    if ((client->writeInt32(owner_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if ((client->write(file_stats->owner,&owner_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    } 
           
    // Write File Group
    int group_length = strlen(file_stats->group) + 1;
    if ((client->writeInt32(group_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if ((client->write(file_stats->group,&group_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    } 
    
    return QUANTAnet_parallelTcpClient_c::OK;
}

int readfilecachestat (QUANTAnet_extendedParallelTcpClient_c *client,struct filecachestat* file_stats) {

    // Read Device
    if ((client->readInt32(file_stats->dev))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Read iNode
    if ((client->readInt64(file_stats->ino))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read Mode
    if ((client->readInt32(file_stats->mode))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read Number of Hard Link
    if ((client->readInt32(file_stats->nlink))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read OwnerId
    if ((client->readInt32(file_stats->uid))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read GroupId
    if ((client->readInt32(file_stats->gid))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Read Device Type
    if ((client->readInt32(file_stats->rdev))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read Size
    if ((client->readInt64(file_stats->size))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read Block Size
    if ((client->readInt32(file_stats->blksize))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Read Block Count
    if ((client->readInt64(file_stats->blocks))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Read Last Access Time
    if ((client->readInt32(file_stats->atime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read Last Modification Time
    if ((client->readInt32(file_stats->mtime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }

    // Read Last Change Time
    if ((client->readInt32(file_stats->ctime))== QUANTAnet_parallelTcpClient_c::FAILED){
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    
    // Read File Owner
    int owner_length = 0;
    if ((client->readInt32(owner_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if(owner_length >= 1024){
        // Invalid String length
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if ((client->read(file_stats->owner,&owner_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
           
    // Read File Group
    int group_length = 0;
    if ((client->readInt32(group_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if(group_length >= 1024){
        // Invalid String length
        return QUANTAnet_parallelTcpClient_c::FAILED;
    }
    if ((client->read(file_stats->group,&group_length))== QUANTAnet_parallelTcpClient_c::FAILED) {
        return QUANTAnet_parallelTcpClient_c::FAILED;
    } 

    return QUANTAnet_parallelTcpClient_c::OK;
}
    
#endif
#endif
