#ifndef GLITE_IO_GUID_HXX
#define GLITE_IO_GUID_HXX
#include <stdlib.h>
#include <unistd.h>


class GUID {
public:
    static const char * NULL_GUID;   
    
protected:
   unsigned int    fGUIDIndex;             //!index in the list of GUIDs in TProcessGUID
   unsigned int    fTimeLow;               // 60 bit time, lower 32 bits
   unsigned short  fTimeMid;               // middle 16 time bits
   unsigned short  fTimeHiAndVersion;      // high 12 time bits + 4 GUID version bits
   unsigned char   fClockSeqHiAndReserved; // high 6 clock bits + 2 bits reserved
   unsigned char   fClockSeqLow;           // low 8 clock bits
   unsigned char   fNode[6];               // 6 node id bytes

    struct guid_time_t {
	unsigned int high;
	unsigned low;
    };

   int   CmpTime(guid_time_t *t1, guid_time_t *t2);
   void  Format(unsigned short clockseq, guid_time_t ts);
   void  GetNodeIdentifier();
   void  GetCurrentTime(guid_time_t *timestamp);
   void  GetSystemTime(guid_time_t *timestamp);
   void  GetRandomInfo(unsigned char seed[16]);
   void  SetFromString(const char *guid_str);

   unsigned int        Sizeof() const { return 18; }

public:
   GUID();
   GUID(const char *guid_str);
   virtual ~GUID();

   const char  *AsString() const;
   int        Compare(const GUID &u) const;
   unsigned short      Hash() const;
   unsigned int        CrossSumHash() const;
   void         Print() const;
   void         PrintBinary();
   int          GetHostAddress() const;
   void         GetGUID(unsigned char guid[16]) const;
   void         SetGUID(const char *guid_str);
   unsigned int       GetGUIDNumber() const { return fGUIDIndex; }
   void         SetGUIDNumber(unsigned index) { fGUIDIndex = index; }

};
#endif // GLITE_IO_GUID_HXX
