#ifndef LIBALIENIO_H
#define LIBALIENIO_H
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef  FCACHE_DEFINE_ALL
#include <glite/data/io/daemon/fcache.hxx>
#else
#define FCACHE_TYPES_ONLY
#include <glite/data/io/daemon/fcache.hxx>
#undef  FCACHE_TYPES_ONLY
#endif

#include <glite/data/io/daemon/aioerror.hxx>

#define TRANSFERBLOCKSIZE 5000000
#define LOCAL_CACHE 1
#define REGIONAL_CACHE 2
#define MAIN_CACHE 4
#define FORCE_RECACHE 8
#define DEFAULT_CACHEREQUEST (MAIN_CACHE+REGIONAL_CACHE+LOCAL_CACHE)
#define MINCACHELEVEL 0
#define MAXCACHELEVEL 15

extern struct timezone tz;

typedef struct TIOstatistic {
    float average1s;
    unsigned long cntaverage1s;
    struct timeval taverage1s;
    float average10s;
    unsigned long cntaverage10s;
    struct timeval taverage10s;
    float average60s;
    unsigned long cntaverage60s;
    struct timeval taverage60s;
    struct timeval measurementtime;
};


namespace log4cpp{
    class Category;
}

class AlienIOclient {
public:
  AlienIOclient(const char* Hostname, const char* Filename, bool precaching = true, const char* IOlib = "posix");
  ~AlienIOclient();
  void SetEncryption(bool cryptauth, bool cryptdata);
  void SetPort(int32 port);
  void SetCacheOptions(bool precaching, int32 CacheSize, int32 NCacheChunks);
  void SetUser(const char* lusername);  
  void SetProtocol(const char* lprotocol);
  void SetLfn(const char* llfn);
  void SetGuid(const char* guid);
  void SetServerID(char id);
  void SetCacheDirectory(const char* cachedir);
  void SetCacheDirectorySize(int32 inMbyte);
  void SetCacheLevel(int32 level);
  bool SetIOtype(const char* iotype);    
  void SetTotalWritesize(long long size);
  void SetPermissions(int perms);
  char* GetRemoteFilename();
  char* GetGUID();
  char* GetLfn();
  bool SetStreaming(bool stream) { streaming = stream; return streaming;}
  int SetNTcp(int n=1);
  void SetConnectionblock(struct connectionblock* sslcrypt);
  void SetTIOstatisticStructs( struct TIOstatistic* Tioread, struct TIOstatistic* Tiowrite) {
      delete readload;
      delete writeload;
      readload  = Tioread;
      writeload = Tiowrite;
      readload->average1s=0;
      gettimeofday(&readload->taverage1s,&tz);
      readload->cntaverage1s=0;
      readload->average10s=0;
      gettimeofday(&readload->taverage10s,&tz);
      readload->cntaverage10s=0;
      readload->average60s=0;
      gettimeofday(&readload->taverage60s,&tz);
      readload->cntaverage60s=0;
      gettimeofday(&readload->measurementtime,&tz);
      writeload->average1s=0;
      gettimeofday(&writeload->taverage1s,&tz);
      writeload->cntaverage1s=0;
      writeload->average10s=0;
      gettimeofday(&writeload->taverage10s,&tz);
      writeload->cntaverage10s=0;
      writeload->average60s=0;
      gettimeofday(&writeload->taverage60s,&tz);
      writeload->cntaverage60s=0;
      gettimeofday(&readload->measurementtime,&tz);      
  }

  int32 Open(int32 flags);
  int32 Read(void *buf, size_t count,struct iovec* iovector=NULL, size_t *iovectorcount=NULL, bool direct=false);
  int32 Write(void *buf, size_t count);
  int64 Lseek(int64 offset, int32 whence);//should return an int64, as this is the type of the offset
  int32 Unlink();
  int32 Close();
  int64 Size();
  int32 Stat(struct filecachestat * file_stats);
  int32 CopyLocal(const char* localname, int silent = 0 , int progbar = 0);
  int32 GetErrno();

  char* GetErrMessageServer();          //getting the error message coming from the server through the cache
  char* GetErrMessageServerUnlink();    //getting the error message coming from the server to the Unlink() case
  

  const char* GetErrnoText();
  static const char* GetErrnoText(int32 myerrno);
  void PrintErrno();
  void PrintErrno(int32 myerrno);

  float GetReadStat1s()   { return readload->average1s; }
  float GetReadStat10s()   { return readload->average10s; }
  float GetReadStat60s()  { return readload->average60s;}
  float GetWriteStat1s()   { return writeload->average1s; }
  float GetWriteStat10s()   { return writeload->average10s; }
  float GetWriteStat60s()  { return writeload->average60s;}
  void  DumpReadStat() ;
  void  DumpWriteStat() ;
  struct TIOstatistic* GetReadStatistic() { return readload;}
  struct TIOstatistic* GetWriteStatistic() { return writeload;}

  static int CacheLookup(char cacheforwardaddress[CACHEFORWARDADDRESSLENGTH], char guid[40]);
    
protected:
    inline void trace(const char *msg, ...);
    inline void traceDebug(const char *msg, ...);
    inline void traceInfo(const char *msg, ...);
    inline void traceWarning(const char *msg, ...);
    inline void traceError(const char *msg, ...);
    
private:
    int64 offset;
    int32 flags;
    void* cache;
    char hostname[1024];
    char filename[FILENAMELENGTH];
    char remotefilename[FILENAMELENGTH];
    int32 cachechunksize;
    int32 ncachechunks;
    bool caching;
    bool streaming;
    int32 ntcp;
    struct connectionblock sslcrypt;
    int32 remoteport;
    struct TIOstatistic *readload;
    struct TIOstatistic *writeload;
    char cachedirectory[FILENAMELENGTH];
    int32 cachedirectorysize; // in Mbyte
    char ioname[128];
 
  
    /**
     * The Logger
     */
    class log4cpp::Category& _logger;
};
#endif
