#ifndef GLITE_DATA_IO_BASE_POSIXIO_HXX
#define GLITE_DATA_IO_BASE_POSIXIO_HXX

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <utime.h>
#include <dlfcn.h>
#include <string.h>

using namespace std;

#define LIB_PREFIX "lib"
#define LIB_SUFFIX ".so"
#define GLITE_DATA_IO_PROTOCOL_PREFIX LIB_PREFIX "glite_data_io_protocol_"

#define POSIX_IO_LIB   GLITE_DATA_IO_PROTOCOL_PREFIX "posixIO"   LIB_SUFFIX
#define OFFLINE_IO_LIB GLITE_DATA_IO_PROTOCOL_PREFIX "offlineIO" LIB_SUFFIX
#define RFIO_IO_LIB    GLITE_DATA_IO_PROTOCOL_PREFIX "rfioIO"    LIB_SUFFIX
#define SHIFT_LIB      LIB_PREFIX                    "shift"     LIB_SUFFIX
#define DCAP_IO_LIB    GLITE_DATA_IO_PROTOCOL_PREFIX "dcapIO"    LIB_SUFFIX
#define DCAP_LIB       LIB_PREFIX                    "dcap"      LIB_SUFFIX


char* iotypes []    = {"null", "posix"     , "file"      , "lfn"       , "guid",       "rfio"     , "dcap",      "gsidcap",   "END"};
char* iolibs  []    = {NULL  , POSIX_IO_LIB, POSIX_IO_LIB, POSIX_IO_LIB, POSIX_IO_LIB, RFIO_IO_LIB, DCAP_IO_LIB, DCAP_IO_LIB, "END"};
char* iohelplibs [] = {NULL  , NULL        , NULL        , NULL        , NULL,         SHIFT_LIB,   DCAP_LIB,    DCAP_LIB,    "END"};

// offline file access (temporary staging)
const char* defaultstagedir = "/tmp/";
char* off_iotypes  []      = {"null","posix"   ,"file"    ,"lfn"     ,"guid"    ,"rfio"      ,"hpss"      ,"hsi"            ,"http" ,"END"};
char* off_get      []      = { NULL ,"cp"      ,"cp"      ,"cp"      ,"cp"      ,"rfcp"      ,"rfcp"      ,"hsi -q get"     ,"wget -q"}; 
char* off_put      []      = { NULL ,"cp"      ,"cp"      ,"cp"      ,"cp"      ,"rfcp"      ,"rfcp"      ,"hsi -q put"     ,""}; 
char* off_sdsep    []      = { NULL ," "       ," "       ," "       ," "       ," "         ," "         ,":"              ,"-O"};
char* off_mkdir    []      = { NULL ,"mkdir -p","mkdir -p","mkdir -p","mkdir -p","nsmkdir -p","rfmkdir -p","hsi -q mkdir -p",""}; 
char* off_rm       []      = { NULL ,"rm -f"   ,"rm -f"   ,"rm -f"   ,"rm -f"   ,"nsrm -f"   ,"rfrm -r"   ,"hsi -q rm"      ,""};
char* off_mv       []      = { NULL ,"mv"      ,"mv"      ,"mv"      ,"mv"      ,"rfrename"  ,"rfrename"  ,"hsi -q rename"  ,""};
int   off_iopp     []      = {0,     0         ,0         ,0         ,0         ,0           ,0           ,0                ,1}; // set 0, if the protocol has to be removed from the pathname, 1 if it has to stay in the pathname
int   off_lhost    []      = {0,     0         ,0         ,0         ,0         ,0           ,0           ,0                ,0}; // set 1 , if the hostname has to stay in front of the path

char gProtocol[4096];

char* getiolib(const char* type) {
    int loop=0;

    // skip direct I/O mode, if AIO_OFFLINE_ACCESS is set
    if (!getenv("AIO_OFFLINE_ACCESS")) {
      while ( (strcmp(iotypes[loop],"END"))) {
	if ( !(strcmp(iotypes[loop],type)))
	  return iolibs[loop];
	loop++;
      }
    }

    loop=0;
    while ( (strcmp(off_iotypes[loop],"END") ) ) {
        if ( !(strcmp(off_iotypes[loop],type))) {
	    sprintf(gProtocol,"%s",off_iotypes[loop]);
	    return OFFLINE_IO_LIB;
	}
	loop++;
    }
    
    return "IO-Type-does-not-exist";
}

char* getiohelplib(const char* type) {
    int loop=0;
    while ( (strcmp(iotypes[loop],"END"))) {
        if ( !(strcmp(iotypes[loop],type)))
            return iohelplibs[loop];
	loop++;
    }

    loop=0;
    while ( (strcmp(off_iotypes[loop],"END") ) ) {
        if ( !(strcmp(off_iotypes[loop],type)))
    	    return NULL;
	loop++;
    }

    return "IO-Type-does-not-exist";
}

class posixIO {
protected:
private:
  
public:
    posixIO() {fIsOpen = false; fWriteMode = false;sprintf(fStagefilename,"");};
    virtual ~posixIO(){};
    
    virtual int open(const char *pathname, int flags, mode_t mode = 0);
    virtual ssize_t read(int fd, void *buf, size_t count);
    virtual ssize_t write(int fd,  void *buf, size_t count);
    virtual int close(int fd);
    virtual int fstat(int filedes, struct stat *buf);
    virtual off_t lseek(int fildes, off_t offset, int whence);
	virtual int unlink(const char *pathname);
    
    // Support For Large Files
#ifdef _LARGEFILE64_SOURCE
    virtual int open64(const char *pathname, int flags, mode_t mode = 0);
    virtual int fstat64(int fildes, struct stat64 *buf);
    virtual off64_t lseek64(int fildes, off64_t offset, int whence);
#endif //_LARGEFILE64_SOURCE

    virtual void  SetProtocol(const char* protocol){sprintf(fProtocol,"%s",protocol);};
    virtual char* GetProtocol(){return fProtocol;};
    virtual char* GetStageFilename() {return NULL;};
    virtual void  SetStageFilename(const char* filename) {};
    virtual char* GetMssFilename() {return NULL;};
    virtual void  SetMssFilename(const char* filename) {};
    virtual bool  IsOpen() {return fIsOpen;};

    char fProtocol[1024];
    char fMssfilename[4096];
    char fStagefilename[4096]; 
    bool fIsOpen;
    bool fWriteMode;
};

typedef posixIO* create_t();
typedef void destroy_t(posixIO*);

struct posixgenerator {
    create_t* create_posix;
    destroy_t* destroy_posix;
    posixIO* newio;
    void* baseposix;
    void* posix;
    void* ioposix;
};


// filter the protocol from a std. url
void URL_protocol(char* urlprotocol,const char* url) {
  char* protend = strstr(url,"://");
  urlprotocol[0] = 0;
  if (protend) {
    memcpy(urlprotocol,url,protend-url);
    urlprotocol[protend-url] = 0;
  } else {
    // default is file://
    sprintf(urlprotocol,"file");
  }
}

// filter the path from a std. url
void URL_path(char* urlpath, const char* url, int leaveprotocol, int leavehost) {
  char* pathstart = strstr(url,"://");
  urlpath[0] = 0;
  if (leaveprotocol) {
    strcpy(urlpath,url);
    return;
  }
  if (pathstart) {
    if (!leavehost) {
      pathstart = strstr(pathstart+3,"/");
      if (!pathstart) 
	return;
    } else {
      pathstart+=3;
    }
    memcpy(urlpath,pathstart,(url+strlen(url)-pathstart));
    urlpath[url+strlen(url)-pathstart]=0;
  } else {
    strcpy(urlpath,url);
    return;
  }
  char* anchor = strchr(urlpath,'#');
  if (anchor)
    anchor = 0;
  char* qmark = strchr(urlpath,'?');
  if (qmark)
    qmark = 0;
}


struct posixgenerator* getIO(char* protocollib, char* iolib) {

    // load the posix library
    //  printf("Loading libraryis %s %s\n",protocollib, iolib);
    void* baseposix = dlopen(POSIX_IO_LIB, RTLD_LAZY|RTLD_GLOBAL);
    if (!baseposix) {
	fprintf(stderr,"Cannot load library : %s\n", dlerror());
	return NULL;
    }

    void* posix=NULL;

    if (protocollib) {
        posix = dlopen(protocollib, RTLD_LAZY|RTLD_GLOBAL);
	if (!posix) {
	    fprintf(stderr,"Cannot load library : %s\n", dlerror());
	    return NULL;
	}
    }

    void* ioposix=NULL;

    if (iolib) {
      //	printf("%d\n",iolib);
	ioposix  = dlopen(iolib, RTLD_LAZY|RTLD_GLOBAL);
	if (!ioposix) {
	    fprintf(stderr,"Cannot load library : %s\n", dlerror());
	    return NULL;
	}
    }
    
    
    struct posixgenerator* generator = new struct posixgenerator;
    if (!generator)
	return NULL;
    // load the symbols
    create_t* create_posix = (create_t*) dlsym(posix, "create");
    destroy_t* destroy_posix = (destroy_t*) dlsym(posix, "destroy");

    if (!create_posix || !destroy_posix) {
	fprintf(stderr,"Cannot load symbols: %s\n", dlerror());
        return NULL;
    }

    // create an instance of the class
    posixIO* newio = create_posix();
    newio->SetProtocol(gProtocol);
    generator->create_posix = create_posix;
    generator->destroy_posix = destroy_posix;
    generator->newio = newio;
    generator->baseposix = baseposix;
    generator->posix = posix;
    generator->ioposix = ioposix;

    return generator;
}

void dropIO(struct posixgenerator* generator) {
    // destroy the class
    generator->destroy_posix(generator->newio);
    if (generator->baseposix)
	dlclose(generator->baseposix);
    if (generator->posix)
	dlclose(generator->posix);
    if (generator->ioposix)
	dlclose(generator->ioposix);
}

struct posixgenerator* PosixInit(char* url) {
    struct posixgenerator* generator;
    char urlprotocol[4096];
    URL_protocol(urlprotocol,url);
    generator = getIO(getiolib(urlprotocol),getiohelplib(urlprotocol));
    if (!generator) {
      fprintf(stderr, "Cannot load requested libraries!");
    }
    return generator;
}

void PosixExit(struct posixgenerator* generator) {
  if (generator) dropIO(generator);
  generator=NULL;
}
#endif // GLITE_DATA_IO_BASE_POSIXIO_HXX
