/*
 *      Copyright (c) 2004 on behalf of the EU EGEE Project:
 *      The European Organization for Nuclear Research (CERN),
 *      Istituto Nazionale di Fisica Nucleare (INFN), Italy
 *      Datamat Spa, Italy
 *      Centre National de la Recherche Scientifique (CNRS), France
 *      CS Systeme d'Information (CSSI), France
 *      Royal Institute of Technology, Center for Parallel Computers (KTH-PDC), Sweden
 *      Universiteit van Amsterdam (UvA), Netherlands
 *      University of Helsinki (UH.HIP), Finland
 *      University of Bergen (UiB), Norway
 *      Council for the Central Laboratory of the Research Councils (CCLRC), United Kingdom
 *
 *      GLite Data IO Base module: Read Configuration Properties
 *
 *      Authors: Paolo Badino <paolo.badino@cern.ch>
 *      Version info: $Id: $
 *      Release: $Name:  $
 *
 *      Revision history:
 *      $Log: $
 *
 */
 
#ifndef GLITE_READCONFIG_HXX_
#define GLITE_READCONFIG_HXX_

#include <string.h>
#include <fstream>

/**
 * Configuration callback
 * @param name [IN] the name of the property to set
 * @param value [IN] the value of the property to set
 * @param data [IN] application specific data passed to readConfiguration 
 */
typedef void (*config_callback) (const char * name, const char * value, void * data);

/**
 * Read the Plug-in Configuration
 * @param file_name [IN] name of configuration
 * @param cb [IN] configuration callback
 * @param data [IN] application specific data passed to readConfiguration 
 */
inline int readConfiguration(const char * file_name, config_callback cb, void * data){
   
    if (0 == cb){
        // No callback provided
        return -1;
    }

    int result;
    // Read the configuration File
    std::ifstream configFile(file_name);
    if(configFile.is_open()){
        char line[1024]; 
        while(configFile.good()){
            configFile.getline(line,1024);

            // Check Separator
            char * pch = strchr(line,'=');
            if(0 == pch){
                continue;
            }
            *pch = '\0';

            // Get Property Name
            char * name = line;
                                                                                
            // Get Property Value
            char * value = pch + 1;
                                    
            // EatWhites on name
            size_t pos = strspn(name, " \t");
            char * name_begin = name + pos;
            pos = strcspn(name_begin, " \t");
            name_begin[pos] = 0;
                                                                                
            // EatWhites on value
            pos = strspn(value, " \t");
            char * value_begin = value + pos;
                                                                                
            pos = 0;
            char * str = value_begin;
            do{
                // Check Black
                pos = strcspn(str, " \t");
                str += pos;
                if(pos == strlen(str) || 0 == pos){
                    break;
                }
                // Check Whites
                pos = strspn(value_begin, " \t");
                str += pos;
                if(pos == strlen(str) || 0 == pos){
                    break;
                }
            }while(pos == 0);
                                                                                
            value_begin[str - value_begin] = 0;
                                                                                
            // Lookup for the Property Properties using the provided callback
            (*cb)(name_begin, value_begin, data);
        }
        result = 0;
    } else {
        result = -1;
    }
    
    return result;
}

#endif //GLITE_READCONFIG_HXX_

